import os
import webbrowser
# from threading import active_count

import aiohttp
import asyncio
from datetime import datetime, timedelta
import logging
import pandas as pd


class MOEXConnectionError(Exception):
    """Исключение для ошибок подключения к MOEX."""
    pass

class Futures:
    def __init__(self, parent):
        self.parent = parent  # Ссылка на родительский класс, если нужно

    def get_sec_info(self, sec_id):
        '''
        Возвращает sectype, grouptype, assetcode, board, market, engine.
        Записывает данные в словарь self.parent.sec_details
        :param sec_id: str Код инструмента
        :return: dict с ключами sectype, grouptype, assetcode, board, market, engine
        '''
        if not isinstance(sec_id, str):
            print(f"Тикер sec_id должен быть строкой, получен тип: {type(sec_id).__name__}")
            return None
        sec_info = asyncio.run(self.parent.get_instrument_info(sec_id))

        if sec_info[-1] is None:
            # raise ValueError(f"Инструмент с sec_id {sec_id} не найден на Бирже")
            print(f"Инструмент с sec_id {sec_id} не найден на Бирже")
            return None

        print(f'Инструмент {sec_id} найден на Бирже')
        self.parent.sec_details[sec_id] = dict(
            sectype=sec_info[0],
            grouptype=sec_info[1],
            assetcode=sec_info[2],
            board=sec_info[3],
            market=sec_info[4],
            engine=sec_info[5]
        )

        if self.parent.sec_details[sec_id]['sectype'] != 'futures':
            print(f'Но это не фьючерс. Тип инструмента: {self.parent.sec_details[sec_id]["sectype"]}')
            return None

        return self.parent.sec_details[sec_id]

    def get_asset_code(self, sec_id):
        '''
        Возвращает Код базового актива (SECID) для кода фьючерсного контракта
        :param sec_id: str, код фьючерсного контракта
        :return: str, ASSETCODE базового актива
        '''
        sec_info = self.get_sec_info(sec_id)
        if sec_info and sec_info['sectype'] == 'futures':
            return sec_info['assetcode']

        return None

    def get_active_contract(self, asset, date=datetime.today()):
        if not isinstance(asset, str):
            print(f"Код базового актива должен быть строкой, получен тип: {type(asset).__name__}")
            return None
        stat = self.get_history_stat(asset, to_active=False, show_table=False)
        return self._get_active_contract(stat, date=date)

    def get_history_list(self, asset, to_active=True):
        if not isinstance(asset, str):
            print(f"Код базового актива должен быть строкой, получен тип: {type(asset).__name__}")
            return None
        stat = self.get_history_stat(asset, to_active=to_active, show_table=False)
        if stat:
            history_list = [item[0] for item in stat]
            history_list.reverse()
            return history_list

        return []

    def get_n_last_contracts(self, asset, n=4, to_active=True):
        if not isinstance(asset, str):
            print(f"Код базового актива должен быть строкой, получен тип: {type(asset).__name__}")
            return None
        if not isinstance(n, int) or n < 0:
            print(f"Значение n ({n}) должен быть целым, положительным числом. Получен неверный тип ({type(n).__name__}) или отрицательное число")
            return None
        clist = self.get_history_list(asset, to_active=to_active)
        n_last_contracts = clist[-n:]
        return n_last_contracts

    def get_contracts_between(self, asset, from_date=datetime.today()-timedelta(days=365), to_date=datetime.today()):
        if not isinstance(asset, str):
            print(f"Код базового актива должен быть строкой, получен тип: {type(asset).__name__}")
            return None
        from_date = self.parent.validate_date(from_date)
        to_date = self.parent.validate_date(to_date)
        if to_date < from_date:
            print(f"Дата to_date {to_date.date()} должна быть больше или равна from_date {from_date.date()}. Получили {to_date < from_date = }")
            return None
        stat = self.get_history_stat(asset, to_active=False, show_table=False)
        contract_list = self.get_history_list(asset, to_active=False)
        from_date_contract = self._get_active_contract(stat, from_date)
        to_date_contract = self._get_active_contract(stat, to_date)
        if not to_date_contract and from_date_contract and from_date_contract in contract_list: # Дата to_date позж активных контартов
            c_list = contract_list[contract_list.index(from_date_contract):]
            return c_list
        if from_date_contract not in contract_list or to_date_contract not in contract_list:
            print("Не найдено активных контрактов на заданные даты начала/конца интервала.")
            return None
        # Получаем индексы контрактов
        index1 = contract_list.index(from_date_contract)
        index2 = contract_list.index(to_date_contract)

        # Проверяем, что первый контракт раньше второго
        if index1 > index2:
            print("Первый контракт (from_date) находится после второго (to_date) в списке.")
            return None

        # Возвращаем срез списка от второго до первого контракта включительно
        c_list = contract_list[index1:index2 + 1]
        return c_list


    def get_contract_exp_date(self, sec_id):
        asset = self.get_asset_code(sec_id)
        if asset:
            stat = self.get_history_stat(asset, to_active=False, show_table=False)
            if stat:
                exp_date = next((item[3] for item in stat if item[0] == sec_id), None)
                if not exp_date:
                    return None

                return exp_date

    def get_previous_contract_exp_date(self, sec_id):
        asset = self.get_asset_code(sec_id)
        if asset:
            stat = self.get_history_stat(asset, to_active=False, show_table=False)
            if stat:
                index = next((i for i, item in enumerate(stat) if item[0] == sec_id), None)
                if index is None or index + 1 >= len(stat):
                    return None
                return stat[index + 1][3]

    def get_history_stat(self, asset, to_active=True, show_table=True):
        stat = asyncio.run(self._get_futures_stat(asset))
        valid_return = stat.get('series', {}).get('data', [])
        # Биржа отдает в качестве ответа список списков, каждый вложенный список имеет вид:
        # ['secid', 'name', 'start_date', 'expiration_date', 'asset_code', 'underlying_asset', 'is_traded']
        # Возвращаю список, у которого вложенные списки будут такие же, но при условии, что 'start_date' не
        # равно None, 'expiration_date' не равно None, и 'name' содержит свои первые 2 символа только один раз
        # (например если 'name' = "SiZ4SiH5", в список не добавляю).
        # if 'series' in stat and 'data' in stat['series'] and stat['series']['data']:

        if valid_return:
            history_stat = [item for item in stat['series']['data']
                            if item[2] is not None and item[3] is not None and
                            item[1].count(item[1][:2]) == 1]
            if to_active:
                active_contract = self._get_active_contract(history_stat)
                history_list = [item[0] for item in history_stat]
                if active_contract and history_list and (active_contract in history_list):
                    target_index = history_list.index(active_contract)
                    history_stat = history_stat[target_index:]
            if show_table:
                columns = ["secid", "shortname", "startdate", "expdate", "assetcode", "underlyingasset",
                           "is_traded"]
                self._display_table(history_stat, columns, name="exp_dates")
                return None
            else:
                return history_stat
        print(f"Биржа не вернула статистику по коду базового актива {asset}. Проверьте его.")
        return None


    def get_all_active_futures(self, show_table=True):
        '''
        Биржа возвращает в качестве ответа список списков, каждый вложенный список имеет вид:
        ["SECID", "BOARDID", "SHORTNAME", "SECNAME", "PREVSETTLEPRICE", "DECIMALS", "MINSTEP", "LASTTRADEDATE",
        "LASTDELDATE", "SECTYPE", "LATNAME", "ASSETCODE", "PREVOPENPOSITION", "LOTVOLUME", "INITIALMARGIN",
        "HIGHLIMIT", "LOWLIMIT", "STEPPRICE", "LASTSETTLEPRICE", "PREVPRICE", "IMTIME", "BUYSELLFEE", "SCALPERFEE",
        "NEGOTIATEDFEE", "EXERCISEFEE"]

        :param show_table: True - сохранять результат в файле html и открывать его в браузере и ничего не возвращать, 
        или (False) вернуть результат в виде списка списков

        :return: html-файл или список списков
        '''

        stat = asyncio.run(self._get_all_futures())
        valid_return = stat.get('securities', {}).get('data', [])
        if valid_return:
            idx = (0,2,11,7,5,6,17,13,14)
            # active_futures = [[self._to_number(item[id]) for id in idx] for item in stat['securities']['data']]
            active_futures = [[item[id] for id in idx] for item in stat['securities']['data']]

            for futures in active_futures:
                info = asyncio.run(self._get_futures_info(futures[0]))
                '''
                url = f"https://iss.moex.com/iss/securities/{secid}.json?iss.meta=off&iss.only=description"
                "description": {
                    "columns": ["name", "title", "value", "type", "sort_order", "is_hidden", "precision"],
                    "data": [
                        ["SECID", "Краткий код", "SiU3", "string", 1, 0, null],
                        ["NAME", "Наименование серии инструмента", "Фьючерсный контракт Si-9.23", "string", 2, 0, null],
                        ["SHORTNAME", "Краткое наименование контракта", "Si-9.23", "string", 3, 0, null],
                        ["LATNAME", "Английское наименование", "Si-9.23", "string", 4, 1, null],
                        ["DELIVERYTYPE", "Исполнение", "В качестве цены исполнения принимается значение <a href=https:\/\/www.moex.com\/ru\/markets\/currency\/get-fixing.aspx?code=USDFIXME>фиксинга USDFIXME<\/a>, рассчитываемого в соответствии с <a href=http:\/\/fs.moex.com\/files\/3971\/>Методикой<\/a>, умноженное на количество долларов США в Лоте и округленное с точностью до целых по правилам математического округления.", "string", 5, 0, null],
                        ["FRSTTRADE", "Начало обращения", "2021-09-24", "date", 9, 0, null],
                        ["LSTTRADE", "Последний день обращения", "2023-09-21", "date", 10, 0, null],
                        ["LSTDELDATE", "Дата исполнения", "2023-09-21", "date", 11, 0, null],
                        ["ASSETCODE", "Код базового актива", "Si", "string", 20, 0, null],
                        ["EXECTYPE", "Тип контракта", "Расчетный", "string", 21, 0, null],
                        ["LOTSIZE", "Лот", "1000", "number", 22, 0, 0],
                        ["CONTRACTNAME", "Наименование контракта", "Фьючерсный контракт на курс доллар США - российский рубль", "string", 23, 0, null],
                        ["GROUPTYPE", "Группа контрактов", "Валюта", "string", 24, 0, null],
                        ["UNIT", "Котировка", "в рублях за лот", "string", 25, 0, null],
                        ["TYPENAME", "Вид контракта", "Фьючерс", "string", 9000, 0, null],
                        ["GROUP", "Код типа инструмента", "futures_forts", "string", 9001, 1, null],
                        ["TYPE", "Тип бумаги", "futures", "string", 10000, 1, null],
                        ["GROUPNAME", "Типа инструмента", "Фьючерсы", "string", 10011, 1, null]
                '''
                valid_return = info.get('description', {}).get('data', [])
                if valid_return:
                    info = {item[0]: item[2] for item in info['description']['data']}
                    futures.insert(3, info["CONTRACTNAME"])
                    futures.insert(4, info["GROUPTYPE"])
                    futures.insert(5, info["FRSTTRADE"])

            if show_table:
                columns = ["SECID", "SHORTNAME", "ASSETCODE", "CONTRACTNAME", "GROUPTYPE", "FRSTTRADE",
                           "LASTTRADEDATE", "DECIMALS", "MINSTEP", "STEPPRICE", "LOTVOLUME", "INITIALMARGIN"]
                self._display_table(active_futures, columns, "active_futures")
                return None
            else:
                return active_futures
        print(f"Биржа не вернула информацию")
        return None

    @staticmethod
    async def _get_all_futures(session=None):
        url = f"https://iss.moex.com/iss/engines/futures/markets/forts/securities.json?iss.meta=off&nearest=1"
        try:
            async with session or aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    response.raise_for_status()  # Автоматически выбрасывает исключение при ошибке
                    data = await response.json()

                    if not data or 'securities' not in data:
                        raise ValueError(f"Пустой ответ или отсутствуют данные")

                    return data

        except aiohttp.ClientError as e:
            logging.error(f"Ошибка подключения к MOEX: {e}")
            raise MOEXConnectionError(f"Не удалось подключиться к MOEX: {e}")
        except Exception as e:
            logging.error(f"Ошибка получения данных: {e}")
            raise MOEXConnectionError(f"Не удалось получить данные: {e}")

    @staticmethod
    async def _get_futures_info(secid):
            url = f"https://iss.moex.com/iss/securities/{secid}.json?iss.meta=off&iss.only=description"
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(url) as response:
                        response.raise_for_status()
                        data = await response.json()

                        if not data or 'description' not in data:
                            raise ValueError(f"Пустой ответ или отсутствуют данные")

                        return data

            except aiohttp.ClientError as e:
                logging.error(f"Ошибка подключения к MOEX: {e}")
                raise MOEXConnectionError(f"Не удалось подключиться к MOEX: {e}")
            except Exception as e:
                logging.error(f"Ошибка получения данных: {e}")
                raise MOEXConnectionError(f"Не удалось получить данные: {e}")

    async def _get_futures_stat(self, asset, session=None):
        url = f"https://iss.moex.com/iss/statistics/engines/futures/markets/forts/series.json?iss.meta=off&asset_code={asset}&show_expired=1"
        try:
            async with session or aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    response.raise_for_status()  # Автоматически выбрасывает исключение при ошибке
                    data = await response.json()

                    if not data or 'series' not in data:
                        raise ValueError(f"Пустой ответ или отсутствуют данные для актива {asset}")

                    return data

        except aiohttp.ClientError as e:
            logging.error(f"Ошибка подключения к MOEX: {e}")
            raise MOEXConnectionError(f"Не удалось подключиться к MOEX: {e}")
        except Exception as e:
            logging.error(f"Ошибка получения статистики для {asset}: {e}")
            raise MOEXConnectionError(f"Не удалось получить статистику по базовому активу {asset} : {e}")

    def _get_active_contract(self, stat, date=datetime.today()):
        if stat:
            dstat = [item[:2] + [datetime.strptime(item[2], "%Y-%m-%d").date()] +
                         [datetime.strptime(item[3], "%Y-%m-%d").date()] + item[4:]
                         for item in stat]

            data = self.parent.validate_date(date)
            if isinstance(data, datetime):
                data = data.date()
            filtered = [item for item in dstat if item[3] > data]
            if not filtered:
                print(f"Не нашли активных фьючерсных контрактов на дату {data}")
                return None
                # return dstat[0][0]
            closest_item = min(filtered, key=lambda x: x[3] - data)
            return closest_item[0]
        return None

    def _display_table(self, data, columns, name):
        # Создание и отображение таблицы
        df = pd.DataFrame(data, columns=columns).fillna("-")
        styled = df.style.set_properties(**{'text-align': 'center', 'border': '1px solid black'}).set_table_styles(
            [{'selector': 'thead th', 'props': [('border', '1px solid black')]}]
            ).format(formatter=self._to_number).hide(axis="index")

        # Сохранение в HTML
        html_file_path = os.path.abspath(f"{name}.html")
        styled.to_html(html_file_path)
        # df.to_html(html_file_path, index=False, border=1, justify='center')

        # Открытие HTML в браузере
        webbrowser.open(f"file://{html_file_path}")

    @staticmethod
    def _to_number(value):
        try:
            number = float(value)
            if number.is_integer():
                return int(number)
            return number
        except (ValueError, TypeError):
            return value


    asset = get_asset_code
    info = get_sec_info
    stat = get_history_stat
    list = get_history_list
    active = get_active_contract
    nlast = get_n_last_contracts
    expdate = get_contract_exp_date
    prevexpdate = get_previous_contract_exp_date
    all_active = get_all_active_futures
    contracts_between = get_contracts_between


import platform
import aiomoex
import asyncio
from moex_store.dns_client import DNS_ISSClient

# print(f'OS = {platform.system()}')
if platform.system() == "Windows":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

aiomoex.client.ISSClient = DNS_ISSClient



import unittest
from unittest.mock import patch, AsyncMock
import asyncio
from datetime import datetime
from moex_store.store import MoexStore


class TestMoexStore(unittest.TestCase):

    def setUp(self):
        # Создание нового цикла событий перед каждым тестом
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        # Создание экземпляра MoexStore перед каждым тестом
        self.store = MoexStore(write_to_file=False)

    def tearDown(self):
        # Закрытие цикла событий после каждого теста
        self.loop.close()

    @patch('aiohttp.ClientSession.get', new_callable=AsyncMock)
    def test_get_data_success(self, mock_get):
        # Мокируем успешный ответ для _check_connection
        mock_get.return_value.__aenter__.return_value.status = 200
        mock_get.return_value.__aenter__.return_value.json = AsyncMock(return_value={'key': 'value'})

        # Мокируем успешные ответы для get_instrument_info и get_history_intervals
        with patch('moex_store.store.MoexStore.get_instrument_info',
                   new=AsyncMock(return_value=('TQBR', 'shares', 'stock'))):
            with patch('moex_store.store.MoexStore.get_history_intervals', new=AsyncMock(
                    return_value=[{'interval': 1, 'begin': '2000-01-01 00:00:00', 'end': '2024-12-31 00:00:00'}])):
                with patch('aiomoex.get_market_candles', new=AsyncMock(return_value=[
                    {'begin': '2023-01-01 00:00:00', 'open': 1, 'high': 1, 'low': 1, 'close': 1, 'volume': 1}])):
                    fromdate = datetime(2023, 1, 1)
                    todate = datetime(2023, 1, 10)
                    data = self.store.get_data('SBER', fromdate, todate, '1m')
                    self.assertIsNotNone(data)

    @patch('aiohttp.ClientSession.get', new_callable=AsyncMock)
    def test_get_data_invalid_date(self, mock_get):
        # Мокируем успешный ответ для _check_connection
        mock_get.return_value.__aenter__.return_value.status = 200
        mock_get.return_value.__aenter__.return_value.json = AsyncMock(return_value={'key': 'value'})

        fromdate = datetime(2023, 1, 10)
        todate = datetime(2023, 1, 1)
        with self.assertRaises(ValueError):
            self.store.get_data('SBER', fromdate, todate, '1m')

    @patch('aiohttp.ClientSession.get', new_callable=AsyncMock)
    def test_get_data_invalid_tf(self, mock_get):
        # Мокируем успешный ответ для _check_connection
        mock_get.return_value.__aenter__.return_value.status = 200
        mock_get.return_value.__aenter__.return_value.json = AsyncMock(return_value={'key': 'value'})

        fromdate = datetime(2023, 1, 1)
        todate = datetime(2023, 1, 10)
        with self.assertRaises(ValueError):
            self.store.get_data('SBER', fromdate, todate, '2h')

    @patch('aiohttp.ClientSession.get', new_callable=AsyncMock)
    def test_get_data_instrument_not_found(self, mock_get):
        # Мокируем успешный ответ для _check_connection
        mock_get.return_value.__aenter__.return_value.status = 200
        mock_get.return_value.__aenter__.return_value.json = AsyncMock(return_value={'key': 'value'})

        with patch('moex_store.store.MoexStore.get_instrument_info', new=AsyncMock(return_value=None)):
            fromdate = datetime(2023, 1, 1)
            todate = datetime(2023, 1, 10)
            with self.assertRaises(ValueError):
                self.store.get_data('INVALID', fromdate, todate, '1m')


if __name__ == '__main__':
    unittest.main()



import os
import time
from pprint import pprint

import pandas as pd
import backtrader as bt
import aiohttp
import asyncio
import aiomoex
from datetime import datetime
from tqdm.asyncio import tqdm_asyncio
from moex_store.tf import change_tf
import moex_store.patch_aiohttp
import ssl
from aiohttp.client_exceptions import ClientConnectorCertificateError
from ssl import SSLCertVerificationError
from moex_store.futures import Futures

TF = {'1m': 1, '5m': 5, '10m': 10, '15m': 15, '30m': 30, '1h': 60, '1d': 24, '1w': 7, '1M': 31, '1q': 4}

class MoexStore:
    def __init__(self, write_to_file=True, max_retries=3, retry_delay=2):
        self.wtf = write_to_file
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self.sec_details = {}
        self.futures = Futures(self)
        asyncio.run(self._check_connection())

    def apply_ssl_patch(self):
        # Создаем SSL-контекст с отключенной проверкой сертификатов
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Переопределяем оригинальный метод ClientSession
        _original_init = aiohttp.ClientSession.__init__

        def _patched_init(self, *args, **kwargs):
            if 'connector' not in kwargs:
                kwargs['connector'] = aiohttp.TCPConnector(ssl=ssl_context)
            _original_init(self, *args, **kwargs)

        aiohttp.ClientSession.__init__ = _patched_init

    async def _check_connection(self):
        url = f"https://iss.moex.com/iss/engines.json"
        attempts = 0
        ssl_patched = False

        while attempts < self.max_retries:
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(url) as response:
                        _ = await response.json()
                        if response.status == 200:
                            print("Биржа MOEX доступна для запросов")
                            return
                        else:
                            raise ConnectionError(f"Не удалось подключиться к MOEX: статус {response.status}")
            except (ClientConnectorCertificateError, SSLCertVerificationError) as e:
                if not ssl_patched:
                    print(f"SSL verification failed: {e}")
                    print(f'Похоже вы запускаете приложение на Мак ОС, но не воспользовались рекомендацией по '
                          f'установке сертификатов при инсталляции Python, типа: "Congratulations! Python 3.9.0 '
                          f'for macOS 10.9 or later was successfully installed. One more thing: to verify '
                          f'the identity of secure network connections, this Python needs a set of SSL root '
                          f'certificates. You can download and install a current curated set from the Certifi '
                          f'project by double-clicking on the Install Certificates icon in the Finder window. '
                          f'See the ReadMe file for more information."')
                    print("Ищите и запускайте файл 'Install Certificates.command' в папке Python 3.XX. Пока пробую "
                          "отключить проверку сертификатов.")

                    self.apply_ssl_patch()
                    ssl_patched = True  # патч применен
                else:
                    print(f"Попытка {attempts + 1} с отключенной проверкой SSL не удалась: {e}")
                    attempts += 1
                    if attempts < self.max_retries:
                        time.sleep(self.retry_delay)
            except aiohttp.ClientError as e:
                print(f"Попытка {attempts + 1}: Не удалось подключиться к MOEX: {e}")
                attempts += 1
                if attempts < self.max_retries:
                    time.sleep(self.retry_delay)
            except Exception as e:
                raise ConnectionError(f"Не удалось подключиться к MOEX: {e}")

    # def get_data(self, name, sec_id, fromdate, todate, tf):
    def get_data(self, sec_id, fromdate, todate, tf='1h', name=None):
        fromdate = self.validate_date(fromdate)
        todate = self.validate_date(todate)

        # Проверка значений
        self._validate_inputs(sec_id, fromdate, todate, tf, name)

        # Получение исторических котировок
        moex_data = asyncio.run(self._get_candles_history(sec_id, fromdate, todate, tf))
        # Готовим итоговый дата-фрейм для backtrader.cerebro
        moex_df = self.make_df(moex_data, tf, self.sec_details[sec_id]['market'], sec_id)  # формируем файл с историей
        data = bt.feeds.PandasData(dataname=moex_df, fromdate=fromdate, todate=todate, name=name)

        return data

    getdata = get_data

    @staticmethod
    def validate_date(inp_date):
        if isinstance(inp_date, datetime):
            return inp_date
        elif isinstance(inp_date, str):
            for fmt in ('%Y-%m-%d', '%d-%m-%Y'):
                try:
                    return datetime.strptime(inp_date, fmt)
                except ValueError:
                    continue
            raise ValueError(f"Неверный формат даты: {inp_date}. Используйте тип datetime или тип "
                             f"str в формате 'YYYY-MM-DD' или 'DD-MM-YYYY'.")
        else:
            raise ValueError(f"Дата должна быть типа datetime или str, получили тип: {type(inp_date).__name__}, "
                             f"значение: {inp_date}")

    def _validate_inputs(self, sec_id, fromdate, todate, tf, name):
        if not isinstance(name, str):
           raise ValueError(f"Тип имени источника данных должен быть str, получен {type(name).__name__}")
        if fromdate >= todate:
            raise ValueError(f"fromdate ({fromdate}) должен быть меньше (раньше) todate ({todate}), \n"
                             f"для получения котировок за один день используйте, например для 2023-06-20: \n"
                             f"fromdate = '2023-06-20', todate = '2023-06-21'")

        if tf not in TF:
            raise ValueError(
                f"Тайм-фрейм для {sec_id} должен быть одним из списка: {list(TF.keys())}, получен: {tf = }")

        sec_info = asyncio.run(self.get_instrument_info(sec_id))

        if sec_info[-1] is None:
            raise ValueError(f"Инструмент с sec_id {sec_id} не найден на Бирже")

        print(f'Инструмент {sec_id} найден на Бирже')
        self.sec_details[sec_id] = dict(
            sectype=sec_info[0],
            grouptype=sec_info[1],
            assetcode=sec_info[2],
            board=sec_info[3],
            market=sec_info[4],
            engine=sec_info[5]
        )

        # Проверка доступных интервалов котировок get_history_intervals
        interval_data = asyncio.run(self.get_history_intervals(sec_id, self.sec_details[sec_id]['board'],
                                                               self.sec_details[sec_id]['market'],
                                                               self.sec_details[sec_id]['engine']))

        # [{'begin': '2022-05-18 11:45:00', 'end': '2024-03-21 13:59:00', 'interval': 1,
        # 'board_group_id': 45} {'begin': '2022-04-01 00:00:00', 'end': '2024-01-01 00:00:00', 'interval': 4,
        # 'board_group_id': 45}, ... ]

        if interval_data is None:
            raise ValueError(f"На Бирже нет доступных интервалов котировок для инструмента {sec_id}.")

        # Если запрошен тайм-фрейм 5, 15 или 30 мин, то проверяем наличие на Биржи котировок
        # с тайм-фреймом 1 мин, так как из них будут приготовлены котировки для 5, 15 или 30 мин.
        user_tf = 1 if TF[tf] in (5, 15, 30) else TF[tf]
        valid_interval = next((item for item in interval_data if item['interval'] == user_tf), None)

        if not valid_interval:
            raise ValueError(f"Тайм-фрейм {tf} не доступен для инструмента {sec_id}")

        valid_begin = datetime.strptime(valid_interval['begin'], '%Y-%m-%d %H:%M:%S')
        valid_end = datetime.strptime(valid_interval['end'], '%Y-%m-%d %H:%M:%S')

        if fromdate > valid_end:
            raise ValueError(f"fromdate ({fromdate}) для {sec_id} и тайм-фрейма '{tf}' должен быть меньше (раньше) {valid_end}, \n"
                             f"валидный интервал с {valid_interval['begin']} по {valid_interval['end']}")

        if todate < valid_begin:
            raise ValueError(f"todate ({todate}) для {sec_id} и тайм-фрейма '{tf}' должен быть больше (позже) {valid_begin}, \n"
                             f"валидный интервал с {valid_interval['begin']} по {valid_interval['end']}")


    async def get_instrument_info(self, secid):
        async with aiohttp.ClientSession() as session:
            url = f"https://iss.moex.com/iss/securities/{secid}.json"
            # https://iss.moex.com/iss/securities/GZU4.json
            # https://iss.moex.com/iss/engines/futures/markets/forts/securities/RIU4.json
            # https://iss.moex.com/iss/statistics/engines/futures/markets/forts/series.json?asset_code=rts&show_expired=1
            async with session.get(url) as response:
                data = await response.json()

                sectype, grouptype, assetcode, board, market, engine = None, None, None, None, None, None

                if 'description' in data and 'data' in data['description'] and data['description']['data']:
                    description_dict = {item[0]: item[2] for item in data['description']['data']}
                    sectype = description_dict.get("TYPE")
                    grouptype = description_dict.get("GROUPTYPE")
                    assetcode = description_dict.get("ASSETCODE")  # if sectype == "futures" else None

                if 'boards' in data and 'data' in data['boards'] and data['boards']['data']:
                    boards_data = data['boards']['data']
                    columns = data['boards']['columns']

                    # Ищем в data['boards']['data'] строку с is_primary = 1 (это главная доска инструмента)
                    primary_boards = filter(lambda item: dict(zip(columns, item)).get('is_primary') == 1, boards_data)

                    for item in primary_boards:
                        record = dict(zip(columns, item))
                        board = record.get('boardid')
                        market = record.get('market')
                        engine = record.get('engine')

                return sectype, grouptype, assetcode, board, market, engine

    @staticmethod
    async def get_history_intervals(sec_id, board, market, engine):
        async with aiohttp.ClientSession() as session:
            data = await aiomoex.get_market_candle_borders(session, security=sec_id, market=market, engine=engine)
            if data:
                return data

            data = await aiomoex.get_board_candle_borders(session, security=sec_id, board=board, market=market,
                                                          engine=engine)
            if data:
                return data

            return None

    async def _get_candles_history(self, sec_id, fromdate, todate, tf):
        delta = (todate - fromdate).days
        start = fromdate.strftime('%Y-%m-%d %H:%M:%S')
        end = todate.strftime('%Y-%m-%d %H:%M:%S')
        key_tf = tf
        tf = TF[tf]

        if tf in (5, 15, 30):
            resample_tf_value = tf
            tf = 1
        else:
            resample_tf_value = None

        async with aiohttp.ClientSession() as session:
            start_time = time.time()
            if tf in (1, 10, 60, 24):
                estimated_time = self.get_estimated_time(delta, tf)
                print(f'Ожидаемое время загрузки данных (зависит от загрузки серверов MOEX): {estimated_time:.0f} сек.')
                time.sleep(0.1)
                data_task = asyncio.create_task(
                    aiomoex.get_market_candles(session, sec_id, interval=tf, start=start, end=end,
                                               market=self.sec_details[sec_id]['market'],
                                               engine=self.sec_details[sec_id]['engine']))
                pbar_task = asyncio.create_task(
                    self._run_progress_bar(estimated_time, data_task))

                data = await data_task

                await pbar_task
            else:
                print(f'Загружаю котировки ...')
                data = await aiomoex.get_market_candles(session, sec_id, interval=tf, start=start, end=end,
                                                        market=self.sec_details[sec_id]['market'],
                                                        engine=self.sec_details[sec_id]['engine'])

            end_time = time.time()
            elapsed_time = end_time - start_time
            if data:
                print(f'История котировок {sec_id} c {fromdate.strftime("%Y-%m-%d")} по {todate.strftime("%Y-%m-%d")} '
                      f'на тайм-фрейме "{key_tf}" получена за {elapsed_time:.2f} секунды')
            else:
                data = await aiomoex.get_board_candles(session, sec_id, interval=tf, start=start, end=end,
                                                       board=self.sec_details[sec_id]['board'],
                                                       market=self.sec_details[sec_id]['market'],
                                                       engine=self.sec_details[sec_id]['engine'])
                if data:
                    print(f'История котировок для {sec_id} получена с тайм-фреймом {key_tf}')
                else:
                    print(f'История котировок для {sec_id} с тайм-фреймом  {key_tf} не найдена на бирже')
                    return None

            if resample_tf_value:
                tf = resample_tf_value
                print(f'Пересчитываю ТФ для {sec_id} c 1 мин на {tf} мин')
                data = change_tf(data, tf)

            return data

    def make_df(self, data, tf, market, sec_id):
        df = pd.DataFrame(data)
        # print(df.columns)

        # Определим необходимые столбцы
        required_columns = ['open', 'close', 'high', 'low', 'value', 'volume', 'begin']

        # Проверим наличие всех необходимых столбцов
        missing_columns = [col for col in required_columns if col not in df.columns]
        if missing_columns:
            raise KeyError(f"Отсутствуют необходимые столбцы: {', '.join(missing_columns)}")

        if market == 'index':
            df.drop(columns=['volume'], inplace=True)
            df.rename(columns={'value': 'volume'}, inplace=True)  # VOLUME = value, ибо Индексы имеют только value
        else:
            df.drop(columns=['value'], inplace=True)

        df.rename(columns={'begin': 'datetime'}, inplace=True)
        df['datetime'] = pd.to_datetime(df['datetime'])  # Преобразование в datetime
        df = df[['datetime', 'open', 'high', 'low', 'close', 'volume']]
        df.set_index('datetime', inplace=True)

        if self.wtf:
            csv_file_path = f"files_from_moex/{sec_id}_tf-{tf}.csv"
            directory = os.path.dirname(csv_file_path)
            if not os.path.exists(directory):
                os.makedirs(directory)
            try:
                df.to_csv(csv_file_path, sep=',', index=True, header=True)
                print(f'Котировки записаны в файл "{csv_file_path}"')
            except IOError as e:
                print(f"Ошибка при записи файла: {e}")

        return df

    @staticmethod
    def get_estimated_time(delta, tf):
        a, b, c = 0, 0, 0
        if tf == 1:
            a, b, c = 0.0003295019925172705, 0.04689869997675399, 6.337785868761401
        elif tf == 10:
            a, b, c = 4.988531246349836e-06, 0.012451095862652674, 0.48478245834903433
        elif tf in [60, 24]:
            a, b, c = - 1.4234264995077613e-07, 0.0024511947309111748, 0.5573157754716476
        return a * delta ** 2 + b * delta + c

    @staticmethod
    async def _run_progress_bar(duration, data_task):
        with tqdm_asyncio(total=100, desc="Загружаю котировки", leave=True, ncols=100,
                          bar_format='{l_bar}{bar}') as pbar:
            for _ in range(100):
                if data_task.done():
                    pbar.n = 100
                    pbar.refresh()
                    break
                await asyncio.sleep(duration / 100)
                pbar.update(1)


import aiohttp
from aiohttp import client_exceptions
import socket
from aiomoex.client import ISSClient, ISSMoexError


class DNS_ISSClient(ISSClient):
    async def get(self, start=None):
        """Загрузка данных.

        :param start:
            Номер элемента с которого нужно загрузить данные. Используется для дозагрузки данных,
            состоящих из нескольких блоков. При отсутствии данные загружаются с начального элемента.

        :return:
            Блок данных с отброшенной вспомогательной информацией - словарь, каждый ключ которого
            соответствует одной из таблиц с данными. Таблицы являются списками словарей, которые напрямую
            конвертируются в pandas.DataFrame.
        :raises ISSMoexError:
            Ошибка при обращении к ISS Moex.
        """
        url = self._url
        query = self._make_query(start)
        # print(f"{datetime.now()}: DNS_ISSClient.get.url: {url}")
        for dns_server in ['77.88.8.1', '8.8.8.8']:
            connector = None
            try:
                # import ssl  # start
                # ssl_context = ssl.create_default_context()
                # ssl_context.check_hostname = False
                # ssl_context.verify_mode = ssl.CERT_NONE  # end

                resolver = aiohttp.resolver.AsyncResolver(nameservers=[dns_server])
                connector = aiohttp.TCPConnector(resolver=resolver)  # , ssl=ssl_context
                self._session._connector = connector
                async with self._session.get(url, params=query) as response:
                    try:
                        response.raise_for_status()
                    except client_exceptions.ClientResponseError as err:
                        raise ISSMoexError("Неверный url", result.url) from err
                    else:
                        result = await response.json()
                        return result[1]
            except (aiohttp.ClientError, socket.gaierror) as e:
                print(f"Failed to make a request to '{url}' using DNS server '{dns_server}'. Error: {e}")
                continue
            finally:
                if connector and not connector.closed:
                    await connector.close()
        return None

